"""Queries for the selection of measurements on different criteria.

These functions are executing different queries on measurement datasets.
Most of them share the same parameter space with `df` for the DataFrame used
as `pandas` and a `date` object used for context. Some functions might define
additional parameters for context. All of these functions share the same
output type. This allows the chaining of queries.
"""
from datetime import timedelta
from datetime import datetime as dt


def latest_datapoint(df, date):
    """Retrieve the latest measurement not exceding a certain date.

    This function returns the latest measurement, which is still
    before a certain date. It will return a DataFrame which consists of
    a single row. This way it is easier

    :param df: Database with all measurements
    :type df: pandas.DataFrame
    :param date: Upper limit for search of latest measurement
    :type date: datetime.datetime w/ time zone information
    :return: The latest measurement
    :rtype: pandas.DataFrame
    """
    assert date.tzinfo is not None
    df = df[df.index < date]
    result = df[df.index == df.index.max()]
    return result


def last_week(df, date):
    """Retrieve all measurements within a week before a certain date (excluding).

    Returns the measurements of one week earlier than a specific date.
    The date is not included in the query.

    :param df: Database with all measurements
    :type df: pandas.DataFrame
    :param date: Upper limit for search of latest measurement
    :type date: datetime.datetime w/ time zone information
    :return: The latest measurement
    :rtype: pandas.DataFrame
    """
    assert date.tzinfo is not None
    threshold = date - timedelta(days=7)
    result = _windowed_selection(df, threshold, date)
    return result


def last_month(df, date):
    """Retrieve all measurements of last month.

    :param df: Database with all measurements
    :type df: pandas.DataFrame
    :param date: Date used as context to define last month
    :type date: datetime.datetime w/ time zone information
    :return: The latest measurement
    :rtype: pandas.DataFrame
    """
    assert date.tzinfo is not None
    if date.month != 1:
        year, month = (date.year, date.month - 1)
    else:
        year, month = (date.year - 1, 12)
    start = dt(year=year, month=month, day=1, tzinfo=date.tzinfo)
    days_of_month = 33 - (start + timedelta(days=33)).day
    end = start.replace(month=start.month, day=days_of_month, hour=23, minut=59, second=59)
    result = _windowed_selection(df, start, end)
    return result


# Pandas allows for easy selection of month/year by using
# the following syntax: `result = df[df.index.year == date.year - 1]`
# It is very short and nice looking. But(!) it does not consider
# time zone. The following implementation actually considers
# timezones and selects data accordingly.
def last_year(df, date):
    """Retrieve all measurements of last year.

    :param df: Database with all measurements
    :type df: pandas.DataFrame
    :param date: Date used as context to define last year
    :type date: datetime.datetime w/ time zone information
    :return: The latest measurement
    :rtype: pandas.DataFrame
    """
    assert date.tzinfo is not None
    start = dt(year=date.year - 1, month=1, day=1, tzinfo=date.tzinfo)
    end = start.replace(year=start.year + 1) - timedelta(seconds=1)
    result = _windowed_selection(df, start, end)
    return result


def specific_month(df, date, month):
    """Retrieve all measurements of a specific month.

    The month will be calculated relative to the given date. Is the month
    higher than the one defined in the date element, than the measurements
    from last year will be returned. If the month already occurred this year,
    then the data from the current year is returned.

    :param df: Database with all measurements
    :type df: pandas.DataFrame
    :param date: Date used as context to define if month already past this year
    :type date: datetime.datetime w/ time zone information
    :return: The latest measurement
    :rtype: pandas.DataFrame
    """
    assert date.tzinfo is not None
    year = date.year if date.month > month else date.year - 1
    start = dt(year=year, month=month, day=1, tzinfo=date.tzinfo)
    days_of_month = 33 - (start + timedelta(days=33)).day
    end = start.replace(month=start.month, day=days_of_month, hour=23, minut=59, second=59)
    result = _windowed_selection(df, start, end)
    return result


def _windowed_selection(df, start, end):
    """Select a time period in database considering time zones.

    This function takes the timezone into consideration when selecting data.
    Both borders are included in the data selection.

    :param df: Database with all measurements
    :type df: pandas.DataFrame
    :param start: Start date of time period
    :type start: datetime.datetime w/ time zone information
    :param end: End date of time period
    :type end: datetime.datetime w/ time zone information
    """
    return df[(df.index <= end) & (df.index >= start)]



"""Wetter - a command line tool to gather current weather information.

The goal of the application is to provide an easy command line (CLI) tool
for getting information about the current weather situation in Karlsruhe.
It allows for querying for the following data:

- Current/latest measurements
- Measurements from the last week, last month and last year
- Measurements for a specific month (either of this year if the month already
  passed or of last year)

The library is split to three separate submodules.
Each of these submodules are responsible for different aspects of the systems.
The [conn.py](./conn.py) is responsible for the backend.
It handles the connection to the database and unifies the input and output.
The [queries.py](./queries.py) is responsible for querying the
database returned by [conn.py](./conn.py).
It handles the proper selection of the data according to the requested time range.
Finally, the [cli.py](./cli.py) module handles the interface to the user.
It is responsible for parsing the user input and calling the backend.
It additionally handles the formating of the results for a nice view by the user.
"""


def get_version():
    """Return current version of library.

    This uses the version provided/setup in poetry.

    :return: Version number of library
    :rtype: str
    """
    try:
        # try new metadata package
        from importlib import metadata

        return metadata.version("wetter")
    except ImportError:
        # backup routine if Python version <= 3.7
        import pkg_resources

        return pkg_resources.get_distribution("wetter").version


__version__ = get_version()



"""Backend of the wetter library.

This module handles the measurement data.
It is reponsible for defining a unified interface to the local data.
The includes the json structure of the data to be saved on the local machine.
as well as the update of the data using open API services.
The module defines three components:

- `WetterDB`: Read/Write/Update of the database
- `APIForWeatherData`: Definition of an API for access different WetterAPI services
- `QueryTicket`: Common data structure to abstract and regulate query parameters

The `WetterDB` reads and write data to the json backend store.
This can handle all the day to day queries of the cli tool.
If an update is necessary, then the other two components come in play.

The update process itself is a two step process.
First, a `QueryTicket` is defined.
It defines the data to be gathered from outside services e.g. [OpenMeteo](https://open-meteo.com/).
Afterwards this ticket is given to an API service for weather data.
Currently there is the OpenMeteo WeatherAPI service implemented.
Additional service might be added later on.
All API services need to be subclasses of `APIForWeatherData`.
This allows for an easy interchange of APIs.
"""

import json
import os
import time
import pytz
from dataclasses import dataclass
from datetime import datetime as dt
from datetime import timezone as tz

import numpy as np
import pandas as pd
import requests as rqs
from pytz import UTC

# Gather absolute path of json file based on its relative position
DATA_LOCATION = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "data")
DB = os.path.join(DATA_LOCATION, "data.json")

# UTC offset of user timezone
UTC_OFFSET_USER_TZ = -time.altzone if time.daylight else -time.timezone


def utcnow():
    return dt.utcnow().replace(tzinfo=tz.utc)


def now():
    return dt.now().replace(tzinfo=pytz.FixedOffset(UTC_OFFSET_USER_TZ / 60))


def get_db():
    """Return the database with all measurements.

    :return: Database of measurements
    :rtype: WetterDB
    """
    return WetterDB(DB)


class APIForWeatherData:
    """Informal interface to external APIs.

    The informal interface defines a blueprint for the functions to
    be implemented by future APIs. This helps an easy exchange of the
    used API service for the update of the database.

    ¡Caution! This is not a strict interface and will be not enforced.
    But if it looks like a duck, ... :)
    """

    @staticmethod
    def parse(response):
        """Parse and transform the response in a format accepted by WetterDB.

        :param response: Response of API query
        :type response: dict
        :return: Measurements to be added to the WetterDB
        :rtype: pd.DataFrame
        """
        assert type(response) == dict
        raise NotImplementedError("Parsing is not implemented.")

    @staticmethod
    def url(qt):
        """URL of the API to be called.

        The URL will be filled out by the data saved in the QueryTicket.

        :param qt: The request parameters for the API call
        :type qt: QueryTicket
        :return: URL of API an http request should be send to
        :rtype: str
        """
        assert isinstance(qt, QueryTicket), f"Expected Queryticket, got {type(qt)}"
        raise NotImplementedError("URL is not setup.")

    @staticmethod
    def get(qt):
        """Get the data from REST API and parse the result.

        :param qt: The request parameters for the API call
        :type qt: QueryTicket
        :return: HTTP Response of the external API
        :rtype: requests.Response
        """
        assert isinstance(qt, QueryTicket), f"Expected Queryticket, got {type(qt)}"
        raise NotImplementedError("HTTP Get request to API not implemented.")


class OpenMeteoMeasurements(APIForWeatherData):
    @staticmethod
    def parse(response):
        """Parse and tranform the response in a format accepted by WetterDB.

        For details check superclass `APIForWeatherData`.
        """
        assert type(response) == dict
        df = pd.DataFrame(
            {
                "time": [dt.strptime(x, "%Y-%m-%dT%H:%M").astimezone(tz=tz.utc) for x in response["hourly"]["time"]],
                "temperature": response["hourly"]["temperature_2m"],
                "wind": response["hourly"]["windspeed_10m"],
            }
        )
        df = df.set_index("time")
        return df

    @staticmethod
    def url(qt):
        """URL of the API to be called.

        For details check superclass `APIForWeatherData`.
        """
        assert isinstance(qt, QueryTicket), f"Expected Queryticket, got {type(qt)}"
        url = (
            "https://api.open-meteo.com/v1/forecast?"
            + "latitude={lat:.2f}&longitude={lon:.2f}&"
            + "timezone={tz}&start_date={start}&end_date={end}&"
            + "hourly=temperature_2m,windspeed_10m,weathercode&current_weather=true"
        )
        return url.format(
            lat=qt.lat,
            lon=qt.lon,
            tz=qt.tz,
            start=qt.start.strftime("%Y-%m-%d"),
            end=qt.end.strftime("%Y-%m-%d"),
        )

    @staticmethod
    def get(qt):
        """Get the data from REST API and parse the result.

        For details check superclass `APIForWeatherData`.
        """
        assert isinstance(qt, QueryTicket)
        url = OpenMeteoMeasurements.url(qt)
        result = rqs.get(url)
        return result


@dataclass
class QueryTicket:
    """Define, check and possibly restrict request parameters for APIs.

    An instance of `QueryTicket` defines the start and end date, the location
    using lat and lon coordinate, as well as the timezone of the output.
    The provided dataclass already supported defaults for the location
    (i.e. Karlsruhe) and timezone (i.e. UTC).
    Some consistency checks e.g. start date <= end date are done at creation.

    :param start: Start date of the measurements to be queried
    :type start: datetime.date
    :param end: End date of the measurements to be queried
    :type end: datetime.date
    :param lat: Latitude position of the measurements to be queried
    :type lat: float
    :param lon: Longitude position of the measurements to be queried
    :type lon: float
    :param tz: Timezone of the measurements to be queried
    :type tz: str
    """

    start: dt.date
    end: dt.date
    lat: float = 49
    lon: float = 8.41
    tz: str = "UTC"

    def __post_init__(self):
        if self.start > self.end:
            tmp = self.start
            self.start = self.end
            self.end = tmp


class WetterDB:
    """Database of weather measurements for a single location."""

    def __init__(self, filename):
        with open(filename, "r") as f:
            self._raw_data = json.load(f)

        # Setup the measurements (flexibel for future updates)
        number_of_measurements = len(self._raw_data["data"])
        measurements = {self._raw_data["index"][i]: self._raw_data["data"][i] for i in range(0, number_of_measurements)}

        # Setup dataframe
        df = pd.DataFrame({"time": self._raw_data["columns"], **measurements})
        df = df.set_index("time")
        df.index = df.index.astype(np.datetime64).tz_localize("UTC")

        self.df = df
        self.check_df()

    def check_df(self):
        """Check assumptions about the database and its structure/schema.

        :raises: AssertionError
        """
        assert "temperature" in self.df.columns
        assert "wind" in self.df.columns
        assert self.df.ndim == 2
        assert self.df.shape[1] == 2
        assert self.df.index.size > 0
        assert self.df.index[0].tzinfo is not None
        assert self.df.index[0].tzinfo == UTC

    def __getattr__(self, name):
        """Get attribute from inner `pd.DataFrame` if not provided by base class.

        The core compoment of the database is the internal `pd.DataFrame`
        object. This method helps calling the attributes of this core component
        by using the attributes of the WetterDB. First, the attributes of the
        WetterDB are returned. Should this result in an error, then the request
        is delegated to the `pd.DataFrame`. Any error on this level is not catched.

        :param name: Requested attribute name
        :type name: str
        :return: Requested Attribute
        :type: Not clear and based on the request
        :raises: KeyError
        """
        try:
            return self.__dict__[name]
        except KeyError:
            return getattr(self.df, name)

    @staticmethod
    def update(db_location, api=OpenMeteoMeasurements, dry_run=False):
        """Update of a database at a given location.

        :param db_location: Database location
        :type db_location: str
        :param api: API to be used for measurement updates
        :type api: APIForWeatherData
        :param dry_run: Write update on to disk (default) or not
        :type dry_run: bool
        :return: Latest and updated weather measurements
        :rtype: pd.DataFrame
        :raises: Exception (if response status code != 200)
        """
        assert issubclass(api, APIForWeatherData)

        # Build parameters for query
        db = WetterDB(db_location)
        latest = db.df.index.max()
        now = dt.utcnow().astimezone(tz.utc)
        qt = QueryTicket(start=latest, end=now)

        # Request data from API
        resp = api.get(qt)
        if resp.status_code == 200:
            json = resp.json()
            df = api.parse(json)
            df = df[df.index <= qt.end]  # kick out predictions
        else:
            raise Exception(f"An error occurred during request [{resp.status_code}]: {resp.json()}")

        # Merge old and new data as well as eliminate duplicates and prediction data
        # (Some API provide forecast data which are not of interest for us)
        df = pd.concat([db.df, df])
        df = df[~df.index.duplicated(keep="last")]

        if not dry_run:
            df.T.to_json(db_location, orient="split", date_format="iso")
        db.df = df
        return db



"""Command line tool for accessing the wetter library.

Setup of the command line tool (CLI) for the library.
The user should be able to handle all necessary communication with
the backend using this interface incl. updating of the database.
"""

import argparse

from wetter import conn
from wetter import queries as qu

from . import __version__


def main():
    """Parse user query and and pretty print answer from the database."""
    args = parse_args()
    db = conn.get_db()
    now = conn.now()
    latest = qu.latest_datapoint(db.df, now)

    if args.cmd == "update":
        conn.WetterDB.update(conn.DB)
        print("Update successful!")
    elif args.cmd == "compare":
        if args.week:
            average = qu.last_week(db.df, now)
            pretty_print_comparison(latest=latest, average=average, mode="week")
        if args.month:
            average = qu.last_month(db.df, now)
            pretty_print_comparison(latest=latest, average=average, mode="month")
        if args.year:
            average = qu.last_year(db.df, now)
            pretty_print_comparison(latest=latest, average=average, mode="year")
    elif args.cmd == "compare-details":
        average = qu.specific_month(db.df, now, args.month)
        pretty_print_detailed_comparison(average)
    elif args.cmd == "latest":
        pretty_print_latest(latest)
    else:
        raise Exception(f"Can not understand the provided subcommand {args.cmd}")


def get_parser():
    """Define argument parser for CLI tool."""
    parser = argparse.ArgumentParser(
        prog="wetter",
        description="Cli tool to check the outside when you're inside",
        epilog="Have a nice day!",
    )
    parser.add_argument("-v", "--version", action="version", version=__version__)
    subparsers = parser.add_subparsers(help="Subcommands for updates and details", dest="cmd")
    subparsers.add_parser("latest", help="Latest measurment [default]")
    subparsers.add_parser("update", help="Update DB")
    comparison_parser = subparsers.add_parser("compare", help="Compare w/ last week, month, year")
    group = comparison_parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--last-week", action="store_true", dest="week", help="Compare w/ last week")
    group.add_argument("--last-year", action="store_true", dest="year", help="Compare w/ last year")
    group.add_argument("--last-month", action="store_true", dest="month", help="Compare w/ last month")
    detailed_cmp_parser = subparsers.add_parser("compare-details", help="Compare today w/ specific month")
    detailed_cmp_parser.add_argument("month", type=int, choices=range(1, 13), help="Month")
    return parser


def parse_args(args=None):
    """Parse arguments from the command line.

    This is a wrapper around the `argparse` argument parser.
    It is necessary since `argparse` does not allow to setup a default subcommand.
    Should the subcommand be empty, this function will fill in the `latest` subcommand.

    :param args: Arguments to be parsed
    :type args: list
    :return: Parsed arguments
    :rtype: `argparse.Namespace`
    """
    parser = get_parser()
    if args is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(args)
    if args.cmd is None:
        args.cmd = "latest"
    return args


def pretty_print_latest(latest):
    """Pretty print the output of the latest measurement.

    :params latest: Latest measurement from database
    :type latest: pd.DataFrame
    """
    t_now = latest.temperature[0]
    w_now = latest.wind[0]
    msg = f"Currently it is 🌡️ {t_now:.1f}°C and windspeed 🌬️ {w_now:.1f} km/h."
    print(msg)
    # print_disclaimer_latest(latest)


def pretty_print_comparison(latest, average, mode):
    """Pretty print the output of a comparison query.

    :params latest: Latest measurement from database
    :type latest: pd.DataFrame
    :params average: Average measurements of a certain time window
    :type average: pd.DataFrame
    :params mode: Window definition either 'week','month' or 'year'
    :type mode: str
    """
    mode = mode.lower()
    assert mode in ("week", "year", "month")
    variable = "temperature"

    temp_avg = average[variable][0]
    temp_now = latest[variable][0]
    diff = temp_avg - temp_now
    relation = "colder" if diff < 0 else "warmer"
    relation = "same" if diff == 0 else relation

    msg = f"It was on average {diff:.1f}°C {relation} last {mode} ({temp_avg:.1f}°C) then today ({temp_now:.1f}°C)"
    print(msg)
    # print_disclaimer_window(average)


def pretty_print_detailed_comparison(window):
    """Pretty print the output of a detailed month window comparison.

    :params window: Measurement within a certain month
    :type latest: pd.DataFrame
    """
    window.index = window.index.map(lambda x: x.astimezone(conn.now().tzinfo))
    variable = "temperature"
    overall_average = window.mean()[variable]
    daily_average = {data.index[0]: data.mean()[variable] for (day, data) in window.groupby(window.index.day)}
    hotter_days = {k: v for k, v in daily_average.items() if v > overall_average}

    month = window.index[0].strftime("%B %Y")
    msg = f"It was on average 🌡️ {overall_average:.1f}°C in 📅 {month}."
    print(msg)
    print(f"The following {len(hotter_days)} days were hotter 🔥 than the average:")
    for day, temp in hotter_days.items():
        day_str = day.strftime("%Y-%m-%d")
        print(f"{day_str} @ {temp:.1f}°C")
    # print_disclaimer_window(window)


def print_disclaimer_latest(latest):
    """Disclaimer about the data the calculations are based upon.

    :params latest: Latest measurement from database
    :type latest: pd.DataFrame
    """
    latest.index = latest.index.map(lambda x: x.astimezone(conn.now().tzinfo))
    date = latest.index[0].strftime("%Y-%m-%d")
    time = latest.index[0].strftime("%I:%M%p")
    disclaimer = f"Latest measurement on 📅 {date} @ {time}."
    print(disclaimer)


def print_disclaimer_window(window):
    """Disclaimer about the data the calculations are based upon.

    :params window: Measurement within a certain month
    :type latest: pd.DataFrame
    """
    window.index = window.index.map(lambda x: x.astimezone(conn.now().tzinfo))
    num = window.index.size
    start = window.index.min().strftime("%Y-%m-%d")
    end = window.index.max().strftime("%Y-%m-%d")

    disclaimer = f"Average was calculated using #{num} measurements between 📅 {start} - {end}."
    print(disclaimer)


if __name__ == "__main__":
    main()
